from __future__ import annotations

import logging
import sys

import pytest
import pytest_mock

import build._ctx


pytestmark = pytest.mark.contextvars


def test_default_ctx_logger(caplog: pytest.LogCaptureFixture) -> None:
    build._ctx.log('foo')

    [record] = caplog.records
    assert record.name == 'build'
    assert record.levelno == logging.INFO
    assert record.message == 'foo'


def test_ctx_custom_logger(mocker: pytest_mock.MockerFixture) -> None:
    log_stub = mocker.stub('custom_logger')

    build._ctx.LOGGER.set(log_stub)
    build._ctx.log('foo')

    log_stub.assert_called_once_with('foo')


def test_ctx_custom_logger_with_custom_verbosity(mocker: pytest_mock.MockerFixture) -> None:
    log_stub = mocker.stub('custom_logger')

    def log(message: str, **_kwargs: object) -> None:
        if build._ctx.verbosity >= 9000:
            log_stub(message)

    build._ctx.LOGGER.set(log)
    build._ctx.log('foo')
    build._ctx.VERBOSITY.set(9000)
    build._ctx.log('bar')

    log_stub.assert_called_once_with('bar')


@pytest.mark.parametrize(
    ('verbosity', 'kwarg_origins'),
    [
        (0, []),
        (1, [('subprocess', 'cmd'), ('subprocess', 'stdout')]),
    ],
)
def test_custom_subprocess_runner_ctx_logging(
    mocker: pytest_mock.MockerFixture, verbosity: int, kwarg_origins: list[tuple[str, ...]]
) -> None:
    log_stub = mocker.stub('custom_logger')

    build._ctx.LOGGER.set(log_stub)
    build._ctx.VERBOSITY.set(verbosity)

    build._ctx.run_subprocess([sys.executable, '-m', 'build', '-V'])

    assert log_stub.call_count == len(kwarg_origins)
    assert [c.kwargs['kind'] for c in log_stub.call_args_list] == kwarg_origins



from __future__ import annotations

import collections.abc
import os
import typing


__all__ = ['ConfigSettings', 'Distribution', 'StrPath', 'SubprocessRunner']

ConfigSettings = collections.abc.Mapping[str, str | collections.abc.Sequence[str]]
Distribution = typing.Literal['sdist', 'wheel', 'editable']

StrPath = str | os.PathLike[str]

TYPE_CHECKING = False

if TYPE_CHECKING:
    from pyproject_hooks import SubprocessRunner
else:
    SubprocessRunner = collections.abc.Callable[
        [collections.abc.Sequence[str], str | None, collections.abc.Mapping[str, str] | None], None
    ]



# These tests check the sdist, path, and wheel of build to ensure that all are valid.

from __future__ import annotations

import subprocess
import sys
import tarfile
import zipfile

from pathlib import Path, PurePosixPath

import pytest


DIR = Path(__file__).parent.resolve()
MAIN_DIR = DIR.parent


sdist_files = {
    '.dockerignore',
    '.gitignore',
    'CHANGELOG.rst',
    'LICENSE',
    'PKG-INFO',
    'README.md',
    'docs/conf.py',
    'pyproject.toml',
    'src/build/py.typed',
    'tests/constraints.txt',
    'tests/packages/test-cant-build-via-sdist/some-file-that-is-needed-for-build.txt',
    'tests/packages/test-no-project/empty.txt',
    'tests/packages/test-setuptools/MANIFEST.in',
    'tox.toml',
}

sdist_patterns = {
    'docs/*.rst',
    'docs/*/*.rst',
    'docs/_static/*',
    'src/build/*.py',
    'src/build/_compat/*.py',
    'tests/*.py',
    'tests/packages/*/*.py',
    'tests/packages/*/*/*.py',
    'tests/packages/*/pyproject.toml',
    'tests/packages/*/setup.*',
}

sdist_files |= {str(PurePosixPath(p.relative_to(MAIN_DIR))) for path in sdist_patterns for p in MAIN_DIR.glob(path)}

wheel_files = {
    'build/__init__.py',
    'build/__main__.py',
    'build/_builder.py',
    'build/_compat/__init__.py',
    'build/_compat/importlib.py',
    'build/_compat/tarfile.py',
    'build/_compat/tomllib.py',
    'build/_ctx.py',
    'build/_exceptions.py',
    'build/_types.py',
    'build/_util.py',
    'build/env.py',
    'build/py.typed',
    'build/util.py',
    'dist-info/licenses/LICENSE',
    'dist-info/METADATA',
    'dist-info/RECORD',
    'dist-info/WHEEL',
    'dist-info/entry_points.txt',
}


@pytest.mark.network
def test_build_sdist(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.chdir(MAIN_DIR)

    subprocess.run(
        [
            sys.executable,
            '-m',
            'build',
            '--sdist',
            '--outdir',
            str(tmp_path),
        ],
        check=True,
    )

    (sdist,) = tmp_path.glob('*.tar.gz')

    with tarfile.open(str(sdist), 'r:gz') as tar:
        simpler = {n.split('/', 1)[-1] for n in tar.getnames() if '/__pycache__/' not in n}

    assert simpler == sdist_files


@pytest.mark.network
@pytest.mark.parametrize('args', [(), ('--wheel',)], ids=('from_sdist', 'direct'))
def test_build_wheel(monkeypatch: pytest.MonkeyPatch, tmp_path: Path, args: tuple[str, ...]) -> None:
    monkeypatch.chdir(MAIN_DIR)

    subprocess.run(
        [
            sys.executable,
            '-m',
            'build',
            *args,
            '--outdir',
            str(tmp_path),
        ],
        check=True,
    )

    (wheel,) = tmp_path.glob('*.whl')

    with zipfile.ZipFile(str(wheel)) as z:
        names = z.namelist()

    trimmed = {n for n in names if 'dist-info' not in n}
    trimmed |= {f'dist-info/{n.split("/", 1)[-1]}' for n in names if 'dist-info' in n}

    assert trimmed == wheel_files



# SPDX-License-Identifier: MIT

from __future__ import annotations

import contextlib
import contextvars
import importlib.metadata
import os
import os.path
import shutil
import stat
import sys
import sysconfig
import tempfile
import typing

from collections.abc import Callable, Generator
from functools import partial, update_wrapper
from pathlib import Path
from typing import Any

import pytest

import build.env

from build._compat import tomllib


def pytest_addoption(parser: pytest.Parser) -> None:
    os.environ['PYTHONWARNINGS'] = 'ignore:DEPRECATION::pip._internal.cli.base_command'  # for when not run within tox
    os.environ['PIP_DISABLE_PIP_VERSION_CHECK'] = '1'  # do not pollute stderr with upgrade advisory
    parser.addoption('--run-integration', action='store_true', help='run the integration tests')
    parser.addoption('--only-integration', action='store_true', help='only run the integration tests')


PYPY3_WIN_VENV_BAD = (
    sys.implementation.name == 'pypy' and sys.implementation.version < (7, 3, 9) and sys.platform.startswith('win')
)
PYPY3_WIN_M = 'https://foss.heptapod.net/pypy/pypy/-/issues/3323 and https://foss.heptapod.net/pypy/pypy/-/issues/3321'


def pytest_collection_modifyitems(config: pytest.Config, items: list[typing.Any]) -> None:
    skip_int = pytest.mark.skip(reason='integration tests not run (no --run-integration flag)')
    skip_other = pytest.mark.skip(reason='only integration tests are run (got --only-integration flag)')

    if config.getoption('--run-integration') and config.getoption('--only-integration'):  # pragma: no cover
        msg = "--run-integration and --only-integration can't be used together, choose one"
        raise pytest.UsageError(msg)

    if len(items) == 1:  # do not require flags if called directly
        return
    for item in items:
        is_integration_file = is_integration(item)
        if PYPY3_WIN_VENV_BAD and item.get_closest_marker('pypy3323bug') and os.environ.get('PYPY3323BUG', None):
            item.add_marker(pytest.mark.xfail(reason=PYPY3_WIN_M, strict=False))
        if (
            PYPY3_WIN_VENV_BAD
            and item.get_closest_marker('isolated')
            and (
                not (is_integration_file and item.originalname == 'test_build')
                or (hasattr(item, 'callspec') and '--no-isolation' not in item.callspec.params.get('args', []))
            )
        ):
            item.add_marker(pytest.mark.xfail(reason=PYPY3_WIN_M, strict=True))
        if is_integration_file:  # pragma: no cover
            if not config.getoption('--run-integration') and not config.getoption('--only-integration'):
                item.add_marker(skip_int)
        elif config.getoption('--only-integration'):  # pragma: no cover
            item.add_marker(skip_other)
    # run integration tests after unit tests
    items.sort(key=lambda i: 1 if is_integration(i) else 0)


def is_integration(item: pytest.Item) -> bool:
    # item.location is typically a (path, lineno) tuple; cast to str for typing
    return os.path.basename(str(item.location[0])) == 'test_integration.py'


def pytest_runtest_call(item: pytest.Item) -> None:
    if item.get_closest_marker('contextvars'):
        if isinstance(item, pytest.Function):
            wrapped_function = partial(contextvars.copy_context().run, item.obj)
            item.obj = update_wrapper(wrapped_function, item.obj)
        else:
            msg = 'cannot rewrap non-function item'
            raise RuntimeError(msg)


@pytest.fixture
def local_pip(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(build.env._PipBackend, '_has_valid_outer_pip', None)


@pytest.fixture(autouse=True)
def avoid_constraints(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv('PIP_CONSTRAINT', raising=False)
    monkeypatch.delenv('PIP_KEYRING_PROVIDER', raising=False)
    monkeypatch.delenv('UV_CONSTRAINT', raising=False)
    monkeypatch.delenv('UV_KEYRING_PROVIDER', raising=False)


@pytest.fixture(autouse=True)
def _clear_keyring_cache() -> Generator[None, None, None]:
    build.env._has_keyring_cli.cache_clear()
    yield
    build.env._has_keyring_cli.cache_clear()


@pytest.fixture(autouse=True, params=[False])
def has_virtualenv(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch) -> None:
    if request.param is not None:
        monkeypatch.setattr(build.env._PipBackend, '_has_virtualenv', request.param)


@pytest.fixture(scope='session', autouse=True)
def ensure_syconfig_vars_created() -> None:
    # the config vars are globally cached and may use get_path, make sure they are created
    sysconfig.get_config_vars()


@pytest.fixture
def packages_path() -> str:
    return os.path.realpath(os.path.join(__file__, '..', 'packages'))


def is_setuptools(package_path: Path) -> bool:
    if package_path.joinpath('setup.py').is_file():
        return True
    pyproject = package_path / 'pyproject.toml'
    try:
        with pyproject.open('rb') as f:
            pp = tomllib.load(f)
    except (FileNotFoundError, ValueError):
        return True
    return 'setuptools' in pp.get('build-system', {}).get('build-backend', 'setuptools')


def generate_package_path_fixture(package_name: str) -> Callable[..., str]:
    @pytest.fixture
    def fixture(packages_path: str, tmp_path: Path) -> str:
        package_path = Path(packages_path) / package_name
        if not is_setuptools(package_path):
            return str(package_path)

        new_path = tmp_path / package_name
        shutil.copytree(package_path, new_path)
        return str(new_path)

    return fixture


# Generate path fixtures dynamically.
package_names = os.listdir(os.path.join(os.path.dirname(__file__), 'packages'))
for package_name in package_names:
    normalized_name = package_name.replace('-', '_')
    fixture_name = f'package_{normalized_name}'
    globals()[fixture_name] = generate_package_path_fixture(package_name)


@pytest.fixture
def test_no_permission(packages_path: str) -> Generator[str, None, None]:
    path = os.path.join(packages_path, 'test-no-permission')
    file = os.path.join(path, 'pyproject.toml')
    orig_stat = os.stat(file).st_mode

    os.chmod(file, ~stat.S_IRWXU)

    yield os.path.join(packages_path, 'test-no-permission')

    os.chmod(file, orig_stat)


@pytest.fixture
def tmp_dir() -> Generator[str, None, None]:
    path = tempfile.mkdtemp(prefix='python-build-test-')

    yield path

    shutil.rmtree(path)


def pytest_report_header() -> str:
    interesting_packages = [
        'build',
        'colorama',
        'coverage',
        'filelock',
        'packaging',
        'pip',
        'pyproject_hooks',
        'setuptools',
        'tomli',
        'virtualenv',
        'wheel',
    ]
    valid = []
    for package in interesting_packages:
        # Old versions of importlib_metadata made this FileNotFoundError
        with contextlib.suppress(ModuleNotFoundError, FileNotFoundError):
            valid.append(f'{package}=={importlib.metadata.version(package)}')
    reqs = ' '.join(valid)
    return f'installed packages of interest: {reqs}'


@pytest.fixture
def subtests(request: pytest.FixtureRequest) -> Any:
    try:
        return request.getfixturevalue('subtests')
    except pytest.FixtureLookupError:

        class Subtests:
            @staticmethod
            @contextlib.contextmanager
            def test(msg: str | None = None, **_kwargs: object) -> Generator[None, None, None]:  # noqa: ARG004
                yield

        return Subtests()



# SPDX-License-Identifier: MIT

from __future__ import annotations

import build


def test_version() -> None:
    assert build.__version__


def test_dir() -> None:
    assert set(dir(build)) == set(build.__all__)



# SPDX-License-Identifier: MIT

import pathlib
import textwrap


def get_requires_for_build_wheel(config_settings=None):
    return ['tomli']


def prepare_metadata_for_build_wheel(metadata_directory, config_settings=None):
    import tomli

    with open('pyproject.toml', 'rb') as f:
        metadata = tomli.load(f)

    distinfo = pathlib.Path(
        metadata_directory,
        '{}-{}.dist-info'.format(
            metadata['project']['name'].replace('-', '-'),
            metadata['project']['version'],
        ),
    )
    distinfo.mkdir(parents=True, exist_ok=True)
    distinfo.joinpath('METADATA').write_text(
        textwrap.dedent(
            f"""
            Metadata-Version: 2.2
            Name: {metadata['project']['name']}
            Version: {metadata['project']['version']}
            Summary: {metadata['project']['description']}
            """
        ).strip()
    )
    return distinfo.name


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    raise NotImplementedError


def build_sdist(sdist_directory, config_settings=None):
    raise NotImplementedError



from setuptools import setup


setup(
    name='legacy',
    version='1.0.0',
    author='Filipe Laíns',
    author_email='lains@archlinux.org',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    packages=['legacy'],
)



from __future__ import annotations


TYPE_CHECKING = False

if TYPE_CHECKING:
    import subprocess
    import types


class BuildException(Exception):
    """
    Exception raised by :class:`build.ProjectBuilder`.
    """


class BuildBackendException(Exception):
    """
    Exception raised when a backend operation fails.
    """

    def __init__(
        self,
        exception: Exception,
        description: str | None = None,
        exc_info: tuple[type[BaseException], BaseException, types.TracebackType] | tuple[None, None, None] = (
            None,
            None,
            None,
        ),
    ) -> None:
        super().__init__()
        self.exception = exception
        self.exc_info = exc_info
        self._description = description

    def __str__(self) -> str:
        if self._description:
            return self._description
        return f'Backend operation failed: {self.exception!r}'


class BuildSystemTableValidationError(BuildException):
    """
    Exception raised when the ``[build-system]`` table in pyproject.toml is invalid.
    """

    def __str__(self) -> str:
        return f'Failed to validate `build-system` in pyproject.toml: {self.args[0]}'


class FailedProcessError(Exception):
    """
    Exception raised when a setup or preparation operation fails.
    """

    def __init__(self, exception: subprocess.CalledProcessError, description: str) -> None:
        super().__init__()
        self.exception = exception
        self._description = description

    def __str__(self) -> str:
        return self._description


class TypoWarning(Warning):
    """
    Warning raised when a possible typo is found.
    """



# SPDX-License-Identifier: MIT


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    import os.path
    import zipfile

    from build._compat import tomllib

    with open('pyproject.toml', 'rb') as f:
        metadata = tomllib.load(f)

    wheel_basename = f'{metadata["project"]["name"].replace("-", "_")}-{metadata["project"]["version"]}'
    with zipfile.ZipFile(os.path.join(wheel_directory, f'{wheel_basename}-py3-none-any.whl'), 'w') as wheel:
        wheel.writestr(
            f'{wheel_basename}.dist-info/METADATA',
            f"""\
Metadata-Version: 2.2
Name: {metadata['project']['name']}
Version: {metadata['project']['version']}
Summary: {metadata['project']['description']}
""",
        )

    return wheel.filename



# SPDX-License-Identifier: MIT

from __future__ import annotations

import contextlib
import importlib.util
import os
import os.path
import re
import shutil
import subprocess
import sys
import tarfile
import urllib.request

from pathlib import Path

import filelock
import pytest
import pytest_mock

import build.__main__


IS_WINDOWS = sys.platform.startswith('win')
IS_PYPY = sys.implementation.name == 'pypy'
MISSING_UV = not shutil.which('uv')


INTEGRATION_SOURCES = {
    'dateutil': ('dateutil/dateutil', '2.9.0'),
    'pip': ('pypa/pip', '25.0.1'),
    'Solaar': ('pwr-Solaar/Solaar', '1.1.14'),
    'flit': ('pypa/flit', '3.12.0'),
}

_SDIST = re.compile('.*.tar.gz')
_WHEEL = re.compile('.*.whl')
ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

EXCL = frozenset(['.tox', 'dist', '.git', '__pycache__', '.integration-sources', '.github', 'tests', 'docs'])


def get_project(name: str, tmp_path: Path) -> Path:
    dest = tmp_path / name
    if name == 'build':
        # our own project is available in-source, just ignore development files

        def _ignore_folder(base: str, filenames: list[str]) -> list[str]:
            ignore = [
                n for n in filenames if n in EXCL or n.endswith(('_cache', '.egg-info', '.pyc')) or n.startswith('.coverage')
            ]
            if os.path.basename(base) == ROOT and 'build' in filenames:  # ignore build only at root (our module is build too)
                ignore.append('build')
            return ignore

        shutil.copytree(ROOT, str(dest), ignore=_ignore_folder)
        return dest

    # for other projects download from github and cache it
    tar_store = os.path.join(ROOT, '.integration-sources')
    # Checking with exists is not parallel safe so just ignore,
    # if the creation failed we will have another failure soon
    # that will notify the user.
    with contextlib.suppress(OSError):
        os.makedirs(tar_store)

    github_org_repo, version = INTEGRATION_SOURCES[name]
    tar_filename = f'{name}-{version}.tar.gz'
    tarball = os.path.join(tar_store, tar_filename)
    with filelock.FileLock(os.path.join(tar_store, f'{tar_filename}.lock')):
        if not os.path.exists(tarball):
            url = f'https://github.com/{github_org_repo}/archive/{version}.tar.gz'
            with urllib.request.urlopen(url) as request, open(tarball, 'wb') as file_handler:
                shutil.copyfileobj(request, file_handler)
    with tarfile.open(tarball, 'r:gz') as tar_handler:
        tar_handler.extractall(str(dest))
    return dest / f'{name}-{version}'


@pytest.mark.network
@pytest.mark.parametrize(
    'call',
    [
        None,  # via code
        [sys.executable, '-m', 'build'],  # module
        ['pyproject-build'],  # entrypoint
    ],
    ids=['code', 'module', 'entrypoint'],
)
@pytest.mark.parametrize(
    'args',
    [
        [],
        pytest.param(
            ['--installer', 'uv'],
            marks=pytest.mark.skipif(MISSING_UV, reason='uv executable not found'),
        ),
        ['-x', '--no-isolation'],
    ],
    ids=['isolated_pip', 'isolated_uv', 'no_isolation'],
)
@pytest.mark.parametrize(
    'project',
    [
        'build',
        'pip',
        'dateutil',
        'Solaar',
        'flit',
    ],
)
@pytest.mark.isolated
def test_build(
    request: pytest.FixtureRequest,  # noqa: ARG001
    monkeypatch: pytest.MonkeyPatch,
    project: str,
    args: list[str],
    call: list[str] | None,
    tmp_path: Path,
) -> None:
    if args == ['--installer', 'uv'] and IS_WINDOWS and IS_PYPY:
        pytest.xfail('uv cannot find PyPy executable')
    if project in {'build', 'flit'} and '--no-isolation' in args:
        pytest.xfail(f"can't build {project} without isolation due to missing dependencies")
    if project == 'Solaar' and IS_WINDOWS and IS_PYPY:
        pytest.xfail('Solaar fails building wheels via sdists on Windows on PyPy 3')

    monkeypatch.chdir(tmp_path)
    monkeypatch.setenv('SETUPTOOLS_SCM_PRETEND_VERSION', '0+dummy')  # for the projects that use setuptools_scm

    if call and call[0] == 'pyproject-build':
        exe_name = f'pyproject-build{".exe" if sys.platform.startswith("win") else ""}'
        exe = os.path.join(os.path.dirname(sys.executable), exe_name)
        if os.path.exists(exe):
            call[0] = exe
        else:
            pytest.skip('Running via PYTHONPATH, so the pyproject-build entrypoint is not available')
    path = get_project(project, tmp_path)
    pkgs = tmp_path / 'pkgs'
    args = [str(path), '-o', str(pkgs), *args]

    if call is None:
        build.__main__.main(args)
    else:
        subprocess.check_call(call + args)

    pkg_names = os.listdir(str(pkgs))
    assert list(filter(_SDIST.match, pkg_names))
    assert list(filter(_WHEEL.match, pkg_names))


def test_isolation(tmp_dir: str, package_test_flit: str, mocker: pytest_mock.MockerFixture) -> None:
    if importlib.util.find_spec('flit_core'):
        pytest.xfail('flit_core is available -- we want it missing!')  # pragma: no cover

    mocker.patch('build.__main__._error')

    build.__main__.main([package_test_flit, '-o', tmp_dir, '--no-isolation'])
    build.__main__._error.assert_called_with("Backend 'flit_core.buildapi' is not available.")  # type: ignore[attr-defined]



from __future__ import annotations

import pytest

from build.__main__ import _natural_language_list


def test_natural_language_list() -> None:
    assert _natural_language_list(['one']) == 'one'
    assert _natural_language_list(['one', 'two']) == 'one and two'
    assert _natural_language_list(['one', 'two', 'three']) == 'one, two and three'
    with pytest.raises(IndexError, match='no elements'):
        _natural_language_list([])



# SPDX-License-Identifier: MIT

from __future__ import annotations

import contextlib
import io
import json
import os
import pathlib
import re
import subprocess
import sys
import unittest.mock
import venv

from typing import Any

import pytest
import pytest_mock

import build
import build.__main__


pytestmark = pytest.mark.contextvars


build_open_owner = 'builtins'

cwd = os.getcwd()
out = os.path.join(cwd, 'dist')

ANSI_STRIP = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')


@pytest.mark.parametrize(
    ('cli_args', 'build_args', 'build_kwargs', 'hook'),
    [
        (
            [],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['-n'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': False,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['-s'],
            (cwd, out),
            {
                'distributions': ['sdist'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package',
        ),
        (
            ['-w'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package',
        ),
        (
            ['-s', '-w'],
            (cwd, out),
            {
                'distributions': ['sdist', 'wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package',
        ),
        (
            ['source'],
            ('source', os.path.join('source', 'dist')),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['-o', 'out'],
            (cwd, 'out'),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['source', '-o', 'out'],
            ('source', 'out'),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['-x'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': True,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['--installer', 'uv'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': 'uv',
            },
            'build_package_via_sdist',
        ),
        (
            ['-C--flag1', '-C--flag2'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {'--flag1': '', '--flag2': ''},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['-C--flag=value'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {'--flag': 'value'},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['-C--flag1=value', '-C--flag2=other_value', '-C--flag2=extra_value'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {'--flag1': 'value', '--flag2': ['other_value', 'extra_value']},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['--config-json={"one": 1, "two": [2, 3], "three": {"in": "out"}}'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {'one': 1, 'two': [2, 3], 'three': {'in': 'out'}},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['--config-json', '{"outer": {"inner": {"deeper": 2}}}'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {'outer': {'inner': {'deeper': 2}}},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['--config-json', '{}'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': None,
                'installer': None,
            },
            'build_package_via_sdist',
        ),
        (
            ['--dependency-constraints-txt', 'contraints.txt'],
            (cwd, out),
            {
                'distributions': ['wheel'],
                'config_settings': {},
                'isolation': True,
                'skip_dependency_check': False,
                'dependency_constraints_txt': 'contraints.txt',
                'installer': None,
            },
            'build_package_via_sdist',
        ),
    ],
)
def test_parse_args(
    mocker: pytest_mock.MockerFixture,
    cli_args: list[str],
    build_args: tuple[Any, Any],
    build_kwargs: dict[str, Any],
    hook: str,
) -> None:
    build_package = mocker.patch('build.__main__.build_package', return_value=['something'])
    build_package_via_sdist = mocker.patch('build.__main__.build_package_via_sdist', return_value=['something'])

    build.__main__.main(cli_args)

    if hook == 'build_package':
        build_package.assert_called_with(*build_args, **build_kwargs)
    elif hook == 'build_package_via_sdist':
        build_package_via_sdist.assert_called_with(*build_args, **build_kwargs)
    else:  # pragma: no cover
        msg = f'Unknown hook {hook}'
        raise ValueError(msg)


def test_prog() -> None:
    out = io.StringIO()

    with pytest.raises(SystemExit), contextlib.redirect_stdout(out):
        build.__main__.main(['--help'], prog='something')

    assert out.getvalue().startswith('usage: something [-h]')


def test_version(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(['--version'])
    out, _ = capsys.readouterr()
    assert out.startswith(f'build {build.__version__}')


@pytest.mark.isolated
def test_build_isolated(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    required_cmd = mocker.patch(
        'build.ProjectBuilder.get_requires_for_build',
        side_effect=[
            ['dep1', 'dep2'],
        ],
    )
    mocker.patch('build.__main__._error')
    install = mocker.patch('build.env.DefaultIsolatedEnv.install')

    build.__main__.build_package(package_test_flit, '.', ['sdist'])

    install.assert_any_call({'flit_core >=2,<4'}, _fresh=True)

    required_cmd.assert_called_with('sdist', None)
    install.assert_any_call(['dep1', 'dep2'])

    build_cmd.assert_called_with('sdist', '.', None)


def test_build_no_isolation_check_deps_empty(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    # check_dependencies = []
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=[])

    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False)

    build_cmd.assert_called_with('sdist', '.', None)


def test_build_package_passes_config_settings_to_build(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    build_cmd = mocker.patch(
        'build.__main__._build',
        side_effect=[
            os.path.join('dist', 'test_flit-1.0.0.tar.gz'),
            os.path.join('dist', 'test_flit-1.0.0-py3-none-any.whl'),
        ],
    )
    config_settings = {'--flag': 'value'}

    built = build.__main__.build_package(
        package_test_flit,
        '.',
        ['sdist', 'wheel'],
        config_settings=config_settings,
        isolation=False,
        skip_dependency_check=True,
        dependency_constraints_txt=pathlib.Path('constraints.txt'),
        installer='uv',
    )

    assert built == ['test_flit-1.0.0.tar.gz', 'test_flit-1.0.0-py3-none-any.whl']
    build_cmd.assert_has_calls(
        [
            unittest.mock.call(
                False, package_test_flit, '.', 'sdist', config_settings, True, pathlib.Path('constraints.txt'), 'uv'
            ),
            unittest.mock.call(
                False, package_test_flit, '.', 'wheel', config_settings, True, pathlib.Path('constraints.txt'), 'uv'
            ),
        ]
    )


def test_build_package_via_sdist_passes_config_settings_to_build(mocker: pytest_mock.MockerFixture) -> None:
    build_cmd = mocker.patch(
        'build.__main__._build',
        side_effect=[
            os.path.join('dist', 'demo-1.0.0.tar.gz'),
            os.path.join('dist', 'demo-1.0.0-py3-none-any.whl'),
        ],
    )
    mocker.patch('build.__main__.tempfile.mkdtemp', return_value='temp-sdist-dir')
    mocker.patch('build.__main__.shutil.rmtree')
    tar_open = mocker.patch('build._compat.tarfile.TarFile.open')
    mocker.patch('build.__main__._ctx.log')
    config_settings = {'--flag': 'value'}

    built = build.__main__.build_package_via_sdist(
        'src',
        'dist',
        ['wheel'],
        config_settings=config_settings,
        isolation=False,
        skip_dependency_check=True,
        dependency_constraints_txt=pathlib.Path('constraints.txt'),
        installer='uv',
    )

    assert built == ['demo-1.0.0.tar.gz', 'demo-1.0.0-py3-none-any.whl']
    tar_open.return_value.__enter__.return_value.extractall.assert_called_once_with('temp-sdist-dir')
    build_cmd.assert_has_calls(
        [
            unittest.mock.call(False, 'src', 'dist', 'sdist', config_settings, True, pathlib.Path('constraints.txt'), 'uv'),
            unittest.mock.call(
                False,
                os.path.join('temp-sdist-dir', 'demo-1.0.0'),
                'dist',
                'wheel',
                config_settings,
                True,
                pathlib.Path('constraints.txt'),
                'uv',
            ),
        ]
    )
    build.__main__.shutil.rmtree.assert_called_once_with('temp-sdist-dir', ignore_errors=True)  # type: ignore[attr-defined]


@pytest.mark.parametrize(
    ('missing_deps', 'output'),
    [
        ([('foo',)], '\n\tfoo'),
        ([('foo',), ('bar', 'baz', 'qux')], '\n\tfoo\n\tbar -> baz -> qux'),
    ],
)
def test_build_no_isolation_with_check_deps(
    mocker: pytest_mock.MockerFixture, package_test_flit: str, missing_deps: list[tuple[str, ...]], output: str
) -> None:
    error = mocker.patch('build.__main__._error')
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=missing_deps)

    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False)

    build_cmd.assert_called_with('sdist', '.', None)
    error.assert_called_with('Missing dependencies:' + output)


@pytest.mark.parametrize(
    ('cli_args', 'err_msg'),
    [
        (['-Cone=1', '--config-json={"two": 2}'], 'not allowed with argument'),
        (['--config-json={"two": 2'], 'Invalid JSON in --config-json'),
        (['--config-json=[1]'], '--config-json must contain a JSON object'),
    ],
)
def test_config_json_errors(cli_args: list[str], err_msg: str, capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(cli_args)

    outerr = capsys.readouterr()
    assert err_msg in outerr.out or err_msg in outerr.err


@pytest.mark.isolated
def test_build_raises_build_exception(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    mocker.patch('build.ProjectBuilder.get_requires_for_build', side_effect=build.BuildException)
    mocker.patch('build.env.DefaultIsolatedEnv.install')

    with pytest.raises(build.BuildException):
        build.__main__.build_package(package_test_flit, '.', ['sdist'])


@pytest.mark.isolated
def test_build_raises_build_backend_exception(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    mocker.patch('build.ProjectBuilder.get_requires_for_build', side_effect=build.BuildBackendException(Exception('a')))
    mocker.patch('build.env.DefaultIsolatedEnv.install')

    msg = f"Backend operation failed: Exception('a'{',' if sys.version_info < (3, 7) else ''})"
    with pytest.raises(build.BuildBackendException, match=re.escape(msg)):
        build.__main__.build_package(package_test_flit, '.', ['sdist'])


@pytest.mark.network
@pytest.mark.pypy3323bug
def test_build_package(tmp_dir: str, package_test_setuptools: str) -> None:
    build.__main__.build_package(package_test_setuptools, tmp_dir, ['sdist', 'wheel'])

    assert sorted(os.listdir(tmp_dir)) == [
        'test_setuptools-1.0.0-py3-none-any.whl',
        'test_setuptools-1.0.0.tar.gz',
    ]


@pytest.mark.network
@pytest.mark.pypy3323bug
def test_build_package_via_sdist(tmp_dir: str, package_test_setuptools: str) -> None:
    build.__main__.build_package_via_sdist(package_test_setuptools, tmp_dir, ['wheel'])

    assert sorted(os.listdir(tmp_dir)) == [
        'test_setuptools-1.0.0-py3-none-any.whl',
        'test_setuptools-1.0.0.tar.gz',
    ]


@pytest.mark.pypy3323bug
def test_build_package_via_sdist_incomplete_sdist(tmp_dir: str, package_test_cant_build_via_sdist: str) -> None:
    with pytest.raises(build.BuildBackendException):
        build.__main__.build_package_via_sdist(package_test_cant_build_via_sdist, tmp_dir, ['wheel'])


def test_build_package_via_sdist_invalid_distribution(tmp_dir: str, package_test_setuptools: str) -> None:
    with pytest.raises(ValueError, match='Only binary distributions are allowed but sdist was specified'):
        build.__main__.build_package_via_sdist(package_test_setuptools, tmp_dir, ['sdist'])


@pytest.mark.isolated
def test_build_package_with_constraints(
    mocker: pytest_mock.MockerFixture, tmp_path: pathlib.Path, package_test_flit: str
) -> None:
    install = mocker.patch('build.env.DefaultIsolatedEnv.install')

    constraints_txt_path = tmp_path.joinpath('constraints.txt')
    constraints_txt_path.write_text(
        """\
flit-core==12.34
foo==wot
""",
        encoding='utf-8',
    )

    with pytest.raises(build.BuildBackendException, match=re.escape("Backend 'flit_core.buildapi' is not available.")):
        build.__main__.build_package(package_test_flit, tmp_path, ['wheel'], dependency_constraints_txt=constraints_txt_path)

    install.assert_any_call({'flit_core >=2,<4'}, constraints={'flit-core==12.34', 'foo==wot'}, _fresh=True)


@pytest.mark.pypy3323bug
@pytest.mark.parametrize(
    ('args', 'output'),
    [
        pytest.param(
            [],
            [
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                '* Building wheel from sdist',
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0.tar.gz and test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='via-sdist-isolation',
            marks=[pytest.mark.network, pytest.mark.isolated],
        ),
        pytest.param(
            ['--no-isolation'],
            [
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                '* Building wheel from sdist',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0.tar.gz and test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='via-sdist-no-isolation',
        ),
        pytest.param(
            ['--wheel'],
            [
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='wheel-direct-isolation',
            marks=[pytest.mark.network, pytest.mark.isolated],
        ),
        pytest.param(
            ['--wheel', '--no-isolation'],
            [
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='wheel-direct-no-isolation',
        ),
        pytest.param(
            ['--sdist', '--no-isolation'],
            [
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                'Successfully built test_setuptools-1.0.0.tar.gz',
            ],
            id='sdist-direct-no-isolation',
        ),
        pytest.param(
            ['--sdist', '--wheel', '--no-isolation'],
            [
                '* Getting build dependencies for sdist...',
                '* Building sdist...',
                '* Getting build dependencies for wheel...',
                '* Building wheel...',
                'Successfully built test_setuptools-1.0.0.tar.gz and test_setuptools-1.0.0-py3-none-any.whl',
            ],
            id='sdist-and-wheel-direct-no-isolation',
        ),
        pytest.param(
            ['--metadata'],
            [
                '* Creating isolated environment: venv+pip...',
                '* Getting metadata for wheel...',
                '* Getting build dependencies for wheel...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
            ],
            id='metadata-isolation',
            marks=[pytest.mark.network, pytest.mark.isolated],
        ),
        pytest.param(
            ['--metadata', '--no-isolation'],
            [
                '* Getting build dependencies for wheel...',
                '* Getting metadata for wheel...',
            ],
            id='metadata-no-isolation',
        ),
    ],
)
@pytest.mark.flaky(reruns=5)
def test_logging_output(
    package_test_setuptools: str, tmp_dir: str, capsys: pytest.CaptureFixture[str], args: list[str], output: list[str]
) -> None:
    build.__main__.main([package_test_setuptools, '-o', tmp_dir, *args])
    _, stderr = capsys.readouterr()
    assert set(stderr.splitlines()) <= set(output)


@pytest.mark.pypy3323bug
@pytest.mark.parametrize(
    ('color', 'stderr_error', 'stderr_body'),
    [
        (
            False,
            'ERROR ',
            [
                '* Creating isolated environment: venv+pip...',
                '* Installing packages in isolated environment:',
                '  - setuptools >= 42.0.0',
                '  - this is invalid',
            ],
        ),
        (
            True,
            '\33[91mERROR\33[0m ',
            [
                '\33[1m* Creating isolated environment: venv+pip...\33[0m',
                '\33[1m* Installing packages in isolated environment:\33[0m',
                '  - setuptools >= 42.0.0',
                '  - this is invalid',
            ],
        ),
    ],
    ids=['no-color', 'color'],
)
@pytest.mark.usefixtures('local_pip')
def test_logging_output_env_subprocess_error(
    mocker: pytest_mock.MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    package_test_invalid_requirements: str,
    tmp_dir: str,
    capsys: pytest.CaptureFixture[str],
    color: bool,
    stderr_body: list[str],
    stderr_error: str,
) -> None:
    with contextlib.suppress(ModuleNotFoundError):  # colorama might not be available
        # do not inject hook to have clear output on capsys
        mocker.patch('colorama.init')

    monkeypatch.delenv('NO_COLOR', raising=False)
    monkeypatch.setenv('FORCE_COLOR' if color else 'NO_COLOR', '')

    with pytest.raises(SystemExit):
        build.__main__.main([package_test_invalid_requirements, '-o', tmp_dir])
    outerr = capsys.readouterr()
    stderr = outerr.err.splitlines()

    assert stderr[:4] == stderr_body
    assert stderr[-1].startswith(stderr_error)

    # Newer versions of pip also color stderr - strip them if present
    assert any(ANSI_STRIP.sub('', e).strip().startswith('< ERROR: Invalid requirement: ') for e in stderr)


@pytest.mark.parametrize(
    ('tty', 'env', 'colors'),
    [
        (False, {}, build.__main__._NO_COLORS),
        (True, {}, build.__main__._COLORS),
        (False, {'NO_COLOR': ''}, build.__main__._NO_COLORS),
        (True, {'NO_COLOR': ''}, build.__main__._NO_COLORS),
        (False, {'FORCE_COLOR': ''}, build.__main__._COLORS),
        (True, {'FORCE_COLOR': ''}, build.__main__._COLORS),
    ],
)
def test_colors(
    mocker: pytest_mock.MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
    tty: bool,
    env: dict[str, str],
    colors: dict[str, object],
) -> None:
    mocker.patch('sys.stdout.isatty', return_value=tty)
    for key, value in env.items():
        monkeypatch.setenv(key, value)

    build.__main__._init_colors()

    assert build.__main__._styles.get() == colors


def test_colors_conflict(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setenv('NO_COLOR', '')
        m.setenv('FORCE_COLOR', '')

        with pytest.warns(
            UserWarning,
            match='Both NO_COLOR and FORCE_COLOR environment variables are set, disabling color',
        ):
            build.__main__._init_colors()

        assert build.__main__._styles.get() == build.__main__._NO_COLORS


def test_logging_output_venv_failure(
    monkeypatch: pytest.MonkeyPatch, package_test_flit: str, tmp_dir: str, capsys: pytest.CaptureFixture[str]
) -> None:
    def raise_called_process_err(*_args: object, **_kwargs: object) -> None:
        raise subprocess.CalledProcessError(1, ['test', 'args'], b'stdoutput', b'stderror')

    monkeypatch.setattr(venv.EnvBuilder, 'create', raise_called_process_err)
    monkeypatch.setenv('NO_COLOR', '')

    with pytest.raises(SystemExit):
        build.__main__.main([package_test_flit, '-o', tmp_dir])

    _, stderr = capsys.readouterr()

    assert (
        stderr
        == """\
* Creating isolated environment: venv+pip...
> test args
< stdoutput
< stderror
ERROR Failed to create venv. Maybe try installing virtualenv.
"""
    )


@pytest.mark.contextvars
@pytest.mark.network
def test_verbose_logging_output(
    subtests: Any,
    capfd: pytest.CaptureFixture[str],
    monkeypatch: pytest.MonkeyPatch,
    tmp_dir: str,
    package_test_setuptools: str,
) -> None:
    monkeypatch.setenv('NO_COLOR', '')

    no_of_lines = -1

    for verbosity in range(-2, 3):
        with subtests.test(verbosity=verbosity):
            cmd = [package_test_setuptools, '-w', '-o', tmp_dir]
            if verbosity:
                cmd.insert(0, f'-{("v" if verbosity > 0 else "q") * abs(verbosity)}')

            build.__main__.main(cmd)

            new_no_of_lines = sum(1 for s in capfd.readouterr() for i in s.splitlines())
            assert new_no_of_lines > no_of_lines
            no_of_lines = new_no_of_lines


def test_metadata_json_output(
    capsys: pytest.CaptureFixture[str],
    package_test_setuptools: str,
) -> None:
    build.__main__.main([package_test_setuptools, '--metadata', '-n'])

    stdout = capsys.readouterr().out
    metadata = json.loads(stdout)
    # Name normalised in old versions of setuptools.
    assert metadata['name'] in {'test_setuptools', 'test-setuptools'}
    assert metadata['version'] == '1.0.0'


def test_setup_cli_windows_colorama_available(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('platform.system', return_value='Windows')
    colorama = mocker.MagicMock()
    mocker.patch.dict('sys.modules', {'colorama': colorama})
    build.__main__._setup_cli(verbosity=0)
    colorama.init.assert_called_once()


def test_setup_cli_windows_colorama_missing(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('platform.system', return_value='Windows')
    mocker.patch.dict('sys.modules', {'colorama': None})
    build.__main__._setup_cli(verbosity=0)


def test_setup_cli_non_windows(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('platform.system', return_value='Linux')
    build.__main__._setup_cli(verbosity=0)


def test_metadata_with_distributions_error() -> None:
    with pytest.raises(SystemExit):
        build.__main__.main(['--metadata', '--sdist'])


def test_entrypoint(mocker: pytest_mock.MockerFixture) -> None:
    main = mocker.patch('build.__main__.main')
    build.__main__.entrypoint()
    main.assert_called_once_with(sys.argv[1:])


def test_handle_build_error_build_backend_exception(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('build.__main__._error', side_effect=SystemExit(1))

    exc = ValueError('test error')
    try:
        raise exc
    except ValueError:
        exc_info = sys.exc_info()

    with pytest.raises(SystemExit), build.__main__._handle_build_error():
        raise build.BuildBackendException(exc, exc_info=exc_info)


def test_log_unknown_kind(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('build.__main__._cprint')
    log = build.__main__._make_logger()
    log('message', kind=('unknown',))


def test_build_no_isolation_skip_dependency_check(mocker: pytest_mock.MockerFixture, package_test_flit: str) -> None:
    build_cmd = mocker.patch('build.ProjectBuilder.build', return_value='something')
    build.__main__.build_package(package_test_flit, '.', ['sdist'], isolation=False, skip_dependency_check=True)
    build_cmd.assert_called_with('sdist', '.', None)


def test_build_package_via_sdist_empty_distributions(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch(
        'build.__main__._build',
        return_value=os.path.join('dist', 'demo-1.0.0.tar.gz'),
    )
    result = build.__main__.build_package_via_sdist('src', 'dist', [])
    assert result == ['demo-1.0.0.tar.gz']


def test_parse_config_settings_triple_duplicate() -> None:
    result = build.__main__._parse_config_settings(['--flag=a', '--flag=b', '--flag=c'])
    assert result == {'--flag': ['a', 'b', 'c']}


def test_build_metadata_runner_without_extra_environ(
    mocker: pytest_mock.MockerFixture,
    tmp_path: pathlib.Path,
    package_test_setuptools: str,
) -> None:
    captured_runners: list[Any] = []

    @contextlib.contextmanager
    def fake_bootstrap(*_args: object, **kwargs: object) -> Any:
        captured_runners.append(kwargs['runner'])
        builder = mocker.MagicMock()
        metadata_dir = tmp_path / 'metadata'
        metadata_dir.mkdir()
        (metadata_dir / 'METADATA').write_bytes(b'Metadata-Version: 2.2\nName: test\nVersion: 1.0\n')
        builder.metadata_path.return_value = str(metadata_dir)
        yield builder

    mocker.patch('build.__main__._bootstrap_build_env', side_effect=fake_bootstrap)
    ctx_run = mocker.patch('build.__main__._ctx.run_subprocess')

    build.__main__._build_metadata(package_test_setuptools, '.', ['wheel'], isolation=False, skip_dependency_check=True)

    assert captured_runners
    captured_runners[0](['echo', 'test'])
    ctx_run.assert_called_once_with(['echo', 'test'], None, mocker.ANY)



from __future__ import annotations

import sys
import tarfile


TYPE_CHECKING = False

if TYPE_CHECKING:
    TarFile = tarfile.TarFile

# Per https://peps.python.org/pep-0706/, the "data" filter will become
# the default in Python 3.14. The first series of releases with the filter
# had a broken filter that could not process symlinks correctly.
elif (
    (3, 9, 18) <= sys.version_info < (3, 10)
    or (3, 10, 13) <= sys.version_info < (3, 11)
    or (3, 11, 5) <= sys.version_info < (3, 12)
    or (3, 12) <= sys.version_info < (3, 14)
):

    class TarFile(tarfile.TarFile):  # pragma: no cover
        extraction_filter = staticmethod(tarfile.data_filter)

else:
    TarFile = tarfile.TarFile  # pragma: no cover


__all__ = [
    'TarFile',
]



from __future__ import annotations

import sys


TYPE_CHECKING = False

if TYPE_CHECKING:
    import importlib_metadata as metadata
elif sys.version_info >= (3, 10, 2):
    from importlib import metadata  # pragma: no cover
else:  # pragma: no cover
    try:
        import importlib_metadata as metadata
    except ModuleNotFoundError:
        # helps bootstrapping when dependencies aren't installed
        from importlib import metadata


__all__ = [
    'metadata',
]



# SPDX-License-Identifier: MIT

from __future__ import annotations

import importlib.util
import re
import unittest.mock

import pytest
import pytest_mock

import build.util


@pytest.mark.pypy3323bug
@pytest.mark.parametrize('isolated', [False, pytest.param(True, marks=[pytest.mark.network, pytest.mark.isolated])])
def test_wheel_metadata(package_test_setuptools: str, isolated: bool) -> None:
    metadata = build.util.project_wheel_metadata(package_test_setuptools, isolated)

    # Setuptools < v69.0.3 (https://github.com/pypa/setuptools/pull/4159) normalized this to dashes
    assert metadata['name'].replace('-', '_') == 'test_setuptools'
    assert metadata['version'] == '1.0.0'
    assert isinstance(metadata.json, dict)


@pytest.mark.network
@pytest.mark.pypy3323bug
def test_wheel_metadata_isolation(package_test_flit: str) -> None:
    if importlib.util.find_spec('flit_core'):
        pytest.xfail('flit_core is available -- we want it missing!')  # pragma: no cover

    metadata = build.util.project_wheel_metadata(package_test_flit)

    assert metadata['name'] == 'test_flit'
    assert metadata['version'] == '1.0.0'
    assert isinstance(metadata.json, dict)

    with pytest.raises(
        build.BuildBackendException,
        match=re.escape("Backend 'flit_core.buildapi' is not available."),
    ):
        build.util.project_wheel_metadata(package_test_flit, isolated=False)


@pytest.mark.network
@pytest.mark.pypy3323bug
def test_with_get_requires(package_test_metadata: str) -> None:
    metadata = build.util.project_wheel_metadata(package_test_metadata)

    # Setuptools < v69.0.3 (https://github.com/pypa/setuptools/pull/4159) normalized this to dashes
    assert metadata['name'].replace('-', '_') == 'test_metadata'
    assert str(metadata['version']) == '1.0.0'
    assert metadata['summary'] == 'hello!'
    assert isinstance(metadata.json, dict)


def test_project_wheel_metadata_installs_build_requires_fresh(mocker: pytest_mock.MockerFixture) -> None:
    env = mocker.MagicMock()
    env_cm = mocker.MagicMock()
    env_cm.__enter__.return_value = env
    env_cm.__exit__.return_value = False
    mocker.patch('build.util.DefaultIsolatedEnv', return_value=env_cm)

    builder = mocker.MagicMock()
    builder.build_system_requires = {'dep1'}
    builder.get_requires_for_build.return_value = {'dep2'}
    mocker.patch('build.util.ProjectBuilder.from_isolated_env', return_value=builder)
    metadata = unittest.mock.sentinel.metadata
    mocker.patch('build.util._project_wheel_metadata', return_value=metadata)

    assert build.util.project_wheel_metadata('/tmp/project') is metadata

    assert env.install.call_args_list == [
        mocker.call({'dep1'}, _fresh=True),
        mocker.call({'dep2'}),
    ]



# SPDX-License-Identifier: MIT
from __future__ import annotations

import importlib.util
import logging
import os
import pathlib
import shutil
import subprocess
import sys
import sysconfig
import typing

from pathlib import Path
from types import SimpleNamespace

import pytest
import pytest_mock

from packaging.version import Version

import build
import build.env


IS_PYPY = sys.implementation.name == 'pypy'
IS_WINDOWS = sys.platform.startswith('win')

MISSING_UV = importlib.util.find_spec('uv') is None and not shutil.which('uv')
MISSING_VIRTUALENV = importlib.util.find_spec('virtualenv') is None


def test_make_extra_environ_overrides_pythonpath() -> None:
    with build.env.DefaultIsolatedEnv() as env:
        extra = env.make_extra_environ()
        assert extra['PYTHONPATH'] == ''
        assert env._env_backend.scripts_dir in extra['PATH']


def test_uv_install_strips_pythonpath(
    mocker: pytest_mock.MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv('PYTHONPATH', '/some/leaky/path')
    run_subprocess = mocker.patch('build.env.run_subprocess')
    with build.env.DefaultIsolatedEnv(installer='uv') as env:
        env.install(['some-package'])
    (install_call,) = run_subprocess.call_args_list
    assert 'PYTHONPATH' not in install_call.kwargs['env']


@pytest.mark.isolated
def test_isolation(monkeypatch: pytest.MonkeyPatch) -> None:
    subprocess.check_call([sys.executable, '-c', 'import build.env'])
    # Test that demonstrates the PYTHONPATH leak issue (issue #1047)
    # When PYTHONPATH is set to include build, and the subprocess env
    # is not properly isolated, the import will succeed instead of failing.
    # Only fails on 3.15+ (due to lazy loading)
    monkeypatch.setenv('PYTHONPATH', os.path.dirname(os.path.dirname(os.path.abspath(build.__file__))))
    debug = 'import sys; import os; print(os.linesep.join(sys.path));'
    with build.env.DefaultIsolatedEnv() as env:
        isolated_env = {**os.environ, **env.make_extra_environ()}
        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_call(
                [env.python_executable, '-c', f'{debug} import build.env'],
                env=isolated_env,
            )


@pytest.mark.skipif(IS_PYPY, reason='PyPy3 uses get path to create and provision venv')
@pytest.mark.skipif(sys.platform != 'darwin', reason='workaround for Apple Python')
def test_can_get_venv_paths_with_conflicting_default_scheme(  # pragma: no cover -- skipped on PyPy and non-darwin
    mocker: pytest_mock.MockerFixture,
) -> None:
    get_scheme_names = mocker.patch('sysconfig.get_scheme_names', return_value=('osx_framework_library',))
    with build.env.DefaultIsolatedEnv():
        pass
    assert get_scheme_names.call_count == 1


SCHEME_NAMES = sysconfig.get_scheme_names()


@pytest.mark.skipif('posix_local' not in SCHEME_NAMES, reason='workaround for Debian/Ubuntu Python')
@pytest.mark.skipif('venv' in SCHEME_NAMES, reason='different call if venv is in scheme names')
def test_can_get_venv_paths_with_posix_local_default_scheme(  # pragma: no cover
    mocker: pytest_mock.MockerFixture,
) -> None:
    get_paths = mocker.spy(sysconfig, 'get_paths')
    # We should never call this, but we patch it to ensure failure if we do
    get_default_scheme = mocker.patch('sysconfig.get_default_scheme', return_value='posix_local')
    with build.env.DefaultIsolatedEnv():
        pass
    get_paths.assert_called_once_with(scheme='posix_prefix', vars=mocker.ANY)
    assert get_default_scheme.call_count == 0


def test_venv_executable_missing_post_creation(
    mocker: pytest_mock.MockerFixture,
) -> None:
    venv_create = mocker.patch('venv.EnvBuilder.create')
    with (
        pytest.raises(RuntimeError, match=r'Virtual environment creation failed, executable .* missing'),
        build.env.DefaultIsolatedEnv(),
    ):
        raise AssertionError
    assert venv_create.call_count == 1


@typing.no_type_check
def test_isolated_env_abstract() -> None:
    with pytest.raises(TypeError):
        build.env.IsolatedEnv()

    class PartialEnv(build.env.IsolatedEnv):
        @property
        def executable(self) -> None:
            raise NotImplementedError

    with pytest.raises(TypeError):
        PartialEnv()

    class PartialEnv2(build.env.IsolatedEnv):
        def make_extra_environ(self) -> None:
            raise NotImplementedError

    with pytest.raises(TypeError):
        PartialEnv2()


@pytest.mark.pypy3323bug
def test_isolated_env_log(
    caplog: pytest.LogCaptureFixture,
    mocker: pytest_mock.MockerFixture,
) -> None:
    caplog.set_level(logging.DEBUG)
    mocker.patch('build.env.run_subprocess')

    with build.env.DefaultIsolatedEnv() as env:
        env.install(['something'])

    assert [(record.levelname, record.message) for record in caplog.records] == [
        ('INFO', 'Creating isolated environment: venv+pip...'),
        ('INFO', 'Installing packages in isolated environment:\n- something'),
    ]


@pytest.mark.isolated
@pytest.mark.usefixtures('local_pip')
def test_default_pip_is_never_too_old() -> None:
    with build.env.DefaultIsolatedEnv() as env:
        version = subprocess.check_output(
            [env.python_executable, '-c', 'import pip; print(pip.__version__, end="")'],
            encoding='utf-8',
        )
        assert Version(version) >= Version('19.1')


@pytest.mark.isolated
@pytest.mark.parametrize('pip_version', ['20.2.0', '20.3.0', '21.0.0', '21.0.1'])
@pytest.mark.parametrize('arch', ['x86_64', 'arm64'])
@pytest.mark.usefixtures('local_pip')
def test_pip_needs_upgrade_mac_os_11(
    mocker: pytest_mock.MockerFixture,
    pip_version: str,
    arch: str,
) -> None:
    run_subprocess = mocker.patch('build.env.run_subprocess')
    mocker.patch('platform.system', return_value='Darwin')
    mocker.patch('platform.mac_ver', return_value=('11.0', ('', '', ''), arch))
    mocker.patch('build._compat.importlib.metadata.distributions', return_value=(SimpleNamespace(version=pip_version),))

    min_pip_version = '20.3.0' if arch == 'x86_64' else '21.0.1'

    with build.env.DefaultIsolatedEnv() as env:
        if Version(pip_version) < Version(min_pip_version):
            assert run_subprocess.call_args_list == [
                mocker.call(
                    [env.python_executable, '-Im', 'pip', 'install', '--no-input', f'pip>={min_pip_version}'],
                    env=mocker.ANY,
                ),
                mocker.call(
                    [env.python_executable, '-Im', 'pip', 'uninstall', '--no-input', '-y', 'setuptools'],
                    env=mocker.ANY,
                ),
            ]
        else:
            run_subprocess.assert_called_once_with(
                [env.python_executable, '-Im', 'pip', 'uninstall', '--no-input', '-y', 'setuptools'],
                env=mocker.ANY,
            )


@pytest.mark.parametrize('has_symlink', [True, False] if sys.platform.startswith('win') else [True])
def test_venv_symlink(
    mocker: pytest_mock.MockerFixture,
    has_symlink: bool,
) -> None:
    if has_symlink:
        mocker.patch('os.symlink')
        mocker.patch('os.unlink')
    else:  # pragma: win32 cover
        mocker.patch('os.symlink', side_effect=OSError())

    # Cache must be cleared to rerun
    build.env._fs_supports_symlink.cache_clear()
    supports_symlink = build.env._fs_supports_symlink()
    build.env._fs_supports_symlink.cache_clear()

    assert supports_symlink is has_symlink


def test_install_short_circuits(
    mocker: pytest_mock.MockerFixture,
) -> None:
    with build.env.DefaultIsolatedEnv() as env:
        assert env.path
        install_dependencies = mocker.patch.object(env._env_backend, 'install_dependencies')

        env.install([])
        install_dependencies.assert_not_called()

        env.install(['foo'])
        install_dependencies.assert_called_once()


@pytest.mark.parametrize('verbosity', range(3))
@pytest.mark.parametrize('constraints', [[], ['foo']])
@pytest.mark.parametrize('fresh', [False, True])
@pytest.mark.usefixtures('local_pip')
def test_default_impl_install_cmd_well_formed(
    mocker: pytest_mock.MockerFixture,
    verbosity: int,
    constraints: list[str],
    fresh: bool,
) -> None:
    mocker.patch.object(build.env._ctx, 'verbosity', verbosity)  # type: ignore[attr-defined]

    with build.env.DefaultIsolatedEnv() as env:
        run_subprocess = mocker.patch('build.env.run_subprocess')

        env.install(['some', 'requirements'], constraints, _fresh=fresh)

        run_subprocess.assert_called_once_with(
            [
                env.python_executable,
                '-Im',
                'pip',
                *([f'-{"v" * (verbosity - 1)}'] if verbosity > 1 else []),
                'install',
                *(['--ignore-installed'] if fresh else []),
                '--use-pep517',
                '--no-warn-script-location',
                '--no-compile',
                '--no-input',
                '-r',
                mocker.ANY,
                *(['-c', mocker.ANY] if constraints else []),
            ],
            env=mocker.ANY,
        )


@pytest.mark.parametrize('verbosity', range(3))
@pytest.mark.parametrize('constraints', [[], ['foo']])
@pytest.mark.parametrize('fresh', [False, True])
@pytest.mark.skipif(IS_PYPY, reason='uv cannot find PyPy executable')
@pytest.mark.skipif(MISSING_UV, reason='uv executable not found')
def test_uv_impl_install_cmd_well_formed(  # pragma: no cover -- uv tests are skipped on PyPy, covered on CPython
    mocker: pytest_mock.MockerFixture,
    verbosity: int,
    constraints: list[str],
    fresh: bool,
) -> None:
    mocker.patch.object(build.env._ctx, 'verbosity', verbosity)  # type: ignore[attr-defined]

    with build.env.DefaultIsolatedEnv(installer='uv') as env:
        run_subprocess = mocker.patch('build.env.run_subprocess')

        env.install(['some', 'requirements'], constraints, _fresh=fresh)

        run_subprocess.assert_called_once_with(
            [
                mocker.ANY,
                'pip',
                *(['-vv' if verbosity > 2 else '-v'] if verbosity > 1 else []),
                'install',
                'some',
                'requirements',
                '--python',
                mocker.ANY,
                *(['-c', mocker.ANY] if constraints else []),
            ],
            env=mocker.ANY,
        )
        (install_call,) = run_subprocess.call_args_list
        assert install_call.kwargs['env']['VIRTUAL_ENV'] == env.path


@pytest.mark.usefixtures('local_pip')
@pytest.mark.parametrize(
    ('installer', 'env_backend_display_name', 'has_virtualenv'),
    [
        ('pip', 'venv+pip', False),
        pytest.param(
            'pip',
            'virtualenv+pip',
            True,
            marks=pytest.mark.skipif(MISSING_VIRTUALENV, reason='virtualenv not found'),
        ),
        pytest.param(
            'pip',
            'virtualenv+pip',
            None,
            marks=pytest.mark.skipif(MISSING_VIRTUALENV, reason='virtualenv not found'),
        ),  # Fall-through
        pytest.param(
            'uv',
            'venv+uv',
            None,
            marks=pytest.mark.skipif(MISSING_UV, reason='uv executable not found'),
        ),
    ],
    indirect=('has_virtualenv',),
)
def test_venv_creation(
    installer: build.env.Installer,
    env_backend_display_name: str,
) -> None:
    with build.env.DefaultIsolatedEnv(installer=installer) as env:
        assert env._env_backend.display_name == env_backend_display_name


@pytest.mark.network
@pytest.mark.usefixtures('local_pip')
@pytest.mark.parametrize(
    'installer',
    [
        'pip',
        pytest.param(
            'uv',
            marks=[
                pytest.mark.skipif(MISSING_UV, reason='uv executable not found'),
            ],
        ),
    ],
)
def test_requirement_installation(
    package_test_flit: str,
    installer: build.env.Installer,
) -> None:
    with build.env.DefaultIsolatedEnv(installer=installer) as env:
        env.install([f'test-flit @ {Path(package_test_flit).as_uri()}'])


@pytest.mark.skipif(MISSING_UV, reason='uv executable not found')
def test_external_uv_detection_success(
    caplog: pytest.LogCaptureFixture,
    mocker: pytest_mock.MockerFixture,
) -> None:
    # Ensure INFO logs are captured
    caplog.set_level(logging.INFO)
    mocker.patch.dict(sys.modules, {'uv': None})

    with build.env.DefaultIsolatedEnv(installer='uv'):
        pass

    # Only check that we logged using an external uv binary (do not rely on
    # which() at assertion time because it can find the environment one).
    # And .text is used instead of .records so a failure message is helpful.
    assert 'Using external uv' in caplog.text


def test_external_uv_detection_failure(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch.dict(sys.modules, {'uv': None})
    mocker.patch('shutil.which', return_value=None)

    with pytest.raises(RuntimeError, match='uv executable not found'), build.env.DefaultIsolatedEnv(installer='uv'):
        raise AssertionError


def test_get_minimum_pip_version_non_darwin(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch('platform.system', return_value='Linux')
    assert build.env._PipBackend._get_minimum_pip_version_str() == '19.1.0'


def test_get_minimum_pip_version_old_darwin(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch('platform.system', return_value='Darwin')
    mocker.patch('platform.mac_ver', return_value=('10.15', ('', '', ''), 'x86_64'))
    assert build.env._PipBackend._get_minimum_pip_version_str() == '19.1.0'


@pytest.mark.parametrize(
    ('version', 'has_no_wheel'),
    [
        pytest.param('20.30.0', True, id='old'),
        pytest.param('20.31.0', False, id='new'),
    ],
)
def test_virtualenv_no_wheel_flag(
    mocker: pytest_mock.MockerFixture,
    version: str,
    has_no_wheel: bool,
) -> None:
    mocker.patch.object(build.env._PipBackend, '_has_valid_outer_pip', None)
    mocker.patch.object(build.env._PipBackend, '_has_virtualenv', True)

    mocker.patch('build._compat.importlib.metadata.version', return_value=version)
    cli_run = mocker.patch('virtualenv.cli_run')
    cli_run.return_value = SimpleNamespace(creator=SimpleNamespace(exe=Path('/fake/python'), script_dir=Path('/fake/scripts')))

    backend = build.env._PipBackend()
    backend.create('/some/path')

    call_args = cli_run.call_args[0][0]
    assert ('--no-wheel' in call_args) is has_no_wheel


def test_install_dependencies_with_outer_pip(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch.object(build.env._PipBackend, '_has_valid_outer_pip', True)

    run_subprocess = mocker.patch('build.env.run_subprocess')

    with build.env.DefaultIsolatedEnv() as env:
        env.install(['some-package'])

    cmd = run_subprocess.call_args_list[-1][0][0]
    assert cmd[:4] == [sys.executable, '-m', 'pip', '--python']


def test_find_executable_osx_framework_scheme(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch('sysconfig.get_scheme_names', return_value=('osx_framework_library', 'posix_prefix'))
    get_paths = mocker.patch(
        'sysconfig.get_paths',
        return_value={
            'scripts': '/fake/bin',
            'purelib': '/fake/lib',
        },
    )
    mocker.patch('os.path.exists', return_value=True)

    _exe, scripts, _purelib = build.env._find_executable_and_scripts('/fake/venv')

    get_paths.assert_called_once_with(scheme='posix_prefix', vars=mocker.ANY)
    assert scripts == '/fake/bin'


def test_find_executable_posix_local_scheme(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch('sysconfig.get_scheme_names', return_value=('posix_local', 'posix_prefix'))
    get_paths = mocker.patch(
        'sysconfig.get_paths',
        return_value={
            'scripts': '/fake/bin',
            'purelib': '/fake/lib',
        },
    )
    mocker.patch('os.path.exists', return_value=True)

    _exe, scripts, _purelib = build.env._find_executable_and_scripts('/fake/venv')

    get_paths.assert_called_once_with(scheme='posix_prefix', vars=mocker.ANY)
    assert scripts == '/fake/bin'


def test_find_executable_fallback_scheme(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch('sysconfig.get_scheme_names', return_value=('posix_prefix',))
    get_paths = mocker.patch(
        'sysconfig.get_paths',
        return_value={
            'scripts': '/fake/bin',
            'purelib': '/fake/lib',
        },
    )
    mocker.patch('os.path.exists', return_value=True)

    _exe, scripts, _purelib = build.env._find_executable_and_scripts('/fake/venv')

    get_paths.assert_called_once_with(vars=mocker.ANY)
    assert scripts == '/fake/bin'


def test_has_dependency_missing() -> None:
    assert build.env._has_dependency('nonexistent_package_xyz_123') is None


@pytest.mark.usefixtures('local_pip')
def test_venv_creation_no_setuptools(
    mocker: pytest_mock.MockerFixture,
) -> None:
    original = build.env._has_dependency

    def no_setuptools(name: str, min_ver: str | None = None, /, **kwargs: object) -> object:
        if name == 'setuptools':
            return None
        return original(name, min_ver, **kwargs)

    mocker.patch('build.env._has_dependency', side_effect=no_setuptools)
    run_subprocess = mocker.patch('build.env.run_subprocess')

    with build.env.DefaultIsolatedEnv() as env:
        assert env.python_executable

    assert all('setuptools' not in str(c) for c in run_subprocess.call_args_list)


def test_has_keyring_cli_found(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('shutil.which', return_value='/usr/bin/keyring')
    assert build.env._has_keyring_cli() is True


def test_has_keyring_cli_missing(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('shutil.which', return_value=None)
    assert build.env._has_keyring_cli() is False


def test_pip_env_with_keyring(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('shutil.which', return_value='/usr/bin/keyring')

    result = build.env._pip_env()

    assert result is not None
    assert result['PIP_KEYRING_PROVIDER'] == 'subprocess'


def test_pip_env_without_keyring(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch('shutil.which', return_value=None)
    assert build.env._pip_env() is None


def test_pip_env_respects_existing_env_var(
    mocker: pytest_mock.MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    mocker.patch('shutil.which', return_value='/usr/bin/keyring')
    monkeypatch.setenv('PIP_KEYRING_PROVIDER', 'import')
    assert build.env._pip_env() is None


@pytest.mark.usefixtures('local_pip')
def test_install_dependencies_passes_keyring_env(
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch('shutil.which', return_value='/usr/bin/keyring')

    with build.env.DefaultIsolatedEnv() as env:
        run_subprocess = mocker.patch('build.env.run_subprocess')
        env.install(['some-package'])

    (install_call,) = run_subprocess.call_args_list
    assert install_call.kwargs['env']['PIP_KEYRING_PROVIDER'] == 'subprocess'


@pytest.mark.skipif(IS_PYPY, reason='uv cannot find PyPy executable')
@pytest.mark.skipif(MISSING_UV, reason='uv executable not found')
def test_uv_install_dependencies_passes_keyring_env(  # pragma: no cover -- uv tests are skipped on PyPy
    mocker: pytest_mock.MockerFixture,
) -> None:
    mocker.patch('shutil.which', return_value='/usr/bin/keyring')

    with build.env.DefaultIsolatedEnv(installer='uv') as env:
        run_subprocess = mocker.patch('build.env.run_subprocess')
        env.install(['some-package'])

    (install_call,) = run_subprocess.call_args_list
    assert install_call.kwargs['env']['UV_KEYRING_PROVIDER'] == 'subprocess'


@pytest.mark.skipif(IS_PYPY, reason='uv cannot find PyPy executable')
@pytest.mark.skipif(MISSING_UV, reason='uv executable not found')
def test_uv_install_respects_existing_keyring_env(  # pragma: no cover -- uv tests are skipped on PyPy
    mocker: pytest_mock.MockerFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    mocker.patch('shutil.which', return_value='/usr/bin/keyring')
    monkeypatch.setenv('UV_KEYRING_PROVIDER', 'disabled')

    with build.env.DefaultIsolatedEnv(installer='uv') as env:
        run_subprocess = mocker.patch('build.env.run_subprocess')
        env.install(['some-package'])

    (install_call,) = run_subprocess.call_args_list
    assert install_call.kwargs['env']['UV_KEYRING_PROVIDER'] == 'disabled'


@pytest.mark.network
def test_pythonpath_does_not_interfere_with_outer_pip(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: pathlib.Path,
) -> None:
    flit_core = tmp_path.joinpath('flit_core-0.0.0.dist-info/')
    flit_core.mkdir()

    monkeypatch.setenv('PYTHONPATH', str(tmp_path))

    with build.env.DefaultIsolatedEnv(installer='pip') as env:
        env.install({'flit_core'}, _fresh=True)

        assert subprocess.check_call([env.python_executable, '-c', 'import flit_core']) == 0



# SPDX-License-Identifier: MIT

from __future__ import annotations

import copy
import logging
import os
import pathlib
import sys
import textwrap
import typing

from collections.abc import Callable
from typing import Any, NoReturn

import pyproject_hooks
import pytest
import pytest_mock

import build
import build._builder

from build._compat import importlib as _importlib


build_open_owner = 'builtins'


DEFAULT_BACKEND = {
    'build-backend': 'setuptools.build_meta:__legacy__',
    'requires': ['setuptools >= 40.8.0'],
}


class MockDistribution(_importlib.metadata.Distribution):
    _metadata: str = ''

    def locate_file(self, path: Any) -> Any:  # pragma: no cover  # noqa: ARG002
        return ''

    def read_text(self, filename: str) -> str:
        if filename == 'METADATA':
            return self._metadata
        return ''

    @classmethod
    def from_name(cls, name: str) -> MockDistribution:
        registry: dict[str, type[MockDistribution]] = {
            'extras_dep': ExtraMockDistribution,
            'requireless_dep': RequirelessMockDistribution,
            'recursive_dep': RecursiveMockDistribution,
            'prerelease_dep': PrereleaseMockDistribution,
            'circular_dep': CircularMockDistribution,
            'nested_circular_dep': NestedCircularMockDistribution,
        }
        if (dist_cls := registry.get(name)) is not None:
            return dist_cls()
        raise _importlib.metadata.PackageNotFoundError


class ExtraMockDistribution(MockDistribution):
    _metadata = textwrap.dedent("""\
        Metadata-Version: 2.2
        Name: extras_dep
        Version: 1.0.0
        Provides-Extra: extra-without-associated-deps
        Provides-Extra: extra-with_unmet-deps
        Requires-Dist: unmet_dep; extra == 'extra-with-unmet-deps'
        Provides-Extra: extra-with-met-deps
        Requires-Dist: extras_dep; extra == 'extra-with-met-deps'
        Provides-Extra: recursive-extra-with-unmet-deps
        Requires-Dist: recursive_dep; extra == 'recursive-extra-with-unmet-deps'""")


class RequirelessMockDistribution(MockDistribution):
    _metadata = textwrap.dedent("""\
        Metadata-Version: 2.2
        Name: requireless_dep
        Version: 1.0.0""")


class RecursiveMockDistribution(MockDistribution):
    _metadata = textwrap.dedent("""\
        Metadata-Version: 2.2
        Name: recursive_dep
        Version: 1.0.0
        Requires-Dist: recursive_unmet_dep""")


class PrereleaseMockDistribution(MockDistribution):
    _metadata = textwrap.dedent("""\
        Metadata-Version: 2.2
        Name: prerelease_dep
        Version: 1.0.1a0""")


class CircularMockDistribution(MockDistribution):
    _metadata = textwrap.dedent("""\
        Metadata-Version: 2.2
        Name: circular_dep
        Version: 1.0.0
        Requires-Dist: nested_circular_dep""")


class NestedCircularMockDistribution(MockDistribution):
    _metadata = textwrap.dedent("""\
        Metadata-Version: 2.2
        Name: nested_circular_dep
        Version: 1.0.0
        Requires-Dist: circular_dep""")


@pytest.mark.parametrize(
    ('requirement_string', 'expected'),
    [
        ('extras_dep', None),
        ('missing_dep', ('missing_dep',)),
        ('requireless_dep', None),
        ('extras_dep[undefined_extra]', None),
        # would the wheel builder filter this out?
        ('extras_dep[extra-without-associated-deps]', None),
        (
            'extras_dep[extra-with-unmet-deps]',
            ('extras_dep[extra-with-unmet-deps]', 'unmet_dep; extra == "extra-with-unmet-deps"'),
        ),
        (
            'extras_dep[recursive-extra-with-unmet-deps]',
            (
                'extras_dep[recursive-extra-with-unmet-deps]',
                'recursive_dep; extra == "recursive-extra-with-unmet-deps"',
                'recursive_unmet_dep',
            ),
        ),
        ('extras_dep[extra-with-met-deps]', None),
        ('missing_dep; python_version>"10"', None),
        ('missing_dep; python_version<="1"', None),
        ('missing_dep; python_version>="1"', ('missing_dep; python_version >= "1"',)),
        ('extras_dep == 1.0.0', None),
        ('extras_dep == 2.0.0', ('extras_dep==2.0.0',)),
        ('extras_dep[extra-without-associated-deps] == 1.0.0', None),
        ('extras_dep[extra-without-associated-deps] == 2.0.0', ('extras_dep[extra-without-associated-deps]==2.0.0',)),
        ('prerelease_dep >= 1.0.0', None),
        ('circular_dep', None),
    ],
)
def test_check_dependency(monkeypatch: pytest.MonkeyPatch, requirement_string: str, expected: tuple[str, ...] | None) -> None:
    monkeypatch.setattr(_importlib.metadata, 'Distribution', MockDistribution)
    assert next(build.check_dependency(requirement_string), None) == expected


def test_bad_project(package_test_no_project: str) -> None:
    # Passing a nonexistent project directory
    with pytest.raises(build.BuildException):
        build.ProjectBuilder(os.path.join(package_test_no_project, 'does-not-exist'))
    # Passing a file as a project directory
    with pytest.raises(build.BuildException):
        build.ProjectBuilder(os.path.join(package_test_no_project, 'empty.txt'))
    # Passing a project directory with no pyproject.toml or setup.py
    with pytest.raises(build.BuildException):
        build.ProjectBuilder(package_test_no_project)


def test_init(
    mocker: pytest_mock.MockerFixture,
    package_test_flit: str,
    package_legacy: str,
    package_test_bad_syntax: str,
) -> None:
    mock_buildcaller = mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    # correct flit pyproject.toml
    builder = build.ProjectBuilder(package_test_flit)
    mock_buildcaller.assert_called_with(
        package_test_flit, 'flit_core.buildapi', backend_path=None, python_executable=sys.executable, runner=builder._runner
    )
    mock_buildcaller.reset_mock()

    # custom python
    builder = build.ProjectBuilder(package_test_flit, python_executable='some-python')
    assert builder.python_executable == 'some-python'
    mock_buildcaller.assert_called_with(
        package_test_flit, 'flit_core.buildapi', backend_path=None, python_executable='some-python', runner=builder._runner
    )
    mock_buildcaller.reset_mock()

    # FileNotFoundError
    builder = build.ProjectBuilder(package_legacy)
    mock_buildcaller.assert_called_with(
        package_legacy,
        'setuptools.build_meta:__legacy__',
        backend_path=None,
        python_executable=sys.executable,
        runner=builder._runner,
    )

    # TomlDecodeError
    with pytest.raises(build.BuildException):
        build.ProjectBuilder(package_test_bad_syntax)


@pytest.mark.skipif(sys.platform.startswith('win'), reason="can't correctly set the permissions required for this")
def test_init_permission_error(test_no_permission: str) -> None:  # pragma: win32 no cover
    with pytest.raises(build.BuildException):
        build.ProjectBuilder(test_no_permission)


def test_init_makes_source_dir_absolute(package_test_flit: str) -> None:
    rel_dir = os.path.relpath(package_test_flit, os.getcwd())
    assert not os.path.isabs(rel_dir)
    builder = build.ProjectBuilder(rel_dir)
    assert os.path.isabs(builder.source_dir)


@pytest.mark.parametrize('distribution', ['wheel', 'sdist'])
def test_get_requires_for_build_missing_backend(
    packages_path: str, distribution: typing.Literal['sdist', 'wheel', 'editable']
) -> None:
    bad_backend_path = os.path.join(packages_path, 'test-bad-backend')
    builder = build.ProjectBuilder(bad_backend_path)

    with pytest.raises(build.BuildBackendException):
        builder.get_requires_for_build(distribution)


@pytest.mark.parametrize('distribution', ['wheel', 'sdist'])
def test_get_requires_for_build_missing_optional_hooks(
    package_test_optional_hooks: str, distribution: typing.Literal['sdist', 'wheel', 'editable']
) -> None:
    builder = build.ProjectBuilder(package_test_optional_hooks)

    assert builder.get_requires_for_build(distribution) == set()


@pytest.mark.parametrize('distribution', ['wheel', 'sdist'])
def test_build_missing_backend(
    packages_path: str, distribution: typing.Literal['sdist', 'wheel', 'editable'], tmpdir: str
) -> None:
    bad_backend_path = os.path.join(packages_path, 'test-bad-backend')
    builder = build.ProjectBuilder(bad_backend_path)

    with pytest.raises(build.BuildBackendException):
        builder.build(distribution, str(tmpdir))


def _nothing_installed(name: str) -> NoReturn:
    raise _importlib.metadata.PackageNotFoundError(name)


def test_check_dependencies(
    mocker: pytest_mock.MockerFixture, package_test_flit: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller.get_requires_for_build_sdist')
    mocker.patch('pyproject_hooks.BuildBackendHookCaller.get_requires_for_build_wheel')

    monkeypatch.setattr(_importlib.metadata, 'distribution', _nothing_installed)

    builder = build.ProjectBuilder(package_test_flit)

    side_effects = [
        [],
        ['something'],
        pyproject_hooks.BackendUnavailable,
    ]

    builder._hook.get_requires_for_build_sdist.side_effect = copy.copy(side_effects)  # type: ignore[attr-defined]
    builder._hook.get_requires_for_build_wheel.side_effect = copy.copy(side_effects)  # type: ignore[attr-defined]

    # requires = []
    assert builder.check_dependencies('sdist') == {('flit_core<4,>=2',)}
    assert builder.check_dependencies('wheel') == {('flit_core<4,>=2',)}

    # requires = ['something']
    assert builder.check_dependencies('sdist') == {('flit_core<4,>=2',), ('something',)}
    assert builder.check_dependencies('wheel') == {('flit_core<4,>=2',), ('something',)}

    # BackendUnavailable
    with pytest.raises(build.BuildBackendException):
        builder.check_dependencies('sdist')
    with pytest.raises(build.BuildBackendException):
        not builder.check_dependencies('wheel')


def test_build(mocker: pytest_mock.MockerFixture, package_test_flit: str, tmp_dir: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)

    builder._hook.build_sdist.side_effect = ['dist.tar.gz', Exception]  # type: ignore[attr-defined]
    builder._hook.build_wheel.side_effect = ['dist.whl', Exception]  # type: ignore[attr-defined]

    assert builder.build('sdist', tmp_dir) == os.path.join(tmp_dir, 'dist.tar.gz')
    builder._hook.build_sdist.assert_called_with(tmp_dir, None)  # type: ignore[attr-defined]

    assert builder.build('wheel', tmp_dir) == os.path.join(tmp_dir, 'dist.whl')
    builder._hook.build_wheel.assert_called_with(tmp_dir, None)  # type: ignore[attr-defined]

    with pytest.raises(build.BuildBackendException):
        builder.build('sdist', tmp_dir)

    with pytest.raises(build.BuildBackendException):
        builder.build('wheel', tmp_dir)


def test_default_backend(mocker: pytest_mock.MockerFixture, package_legacy: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_legacy)

    assert builder._build_system == DEFAULT_BACKEND


def test_missing_backend(mocker: pytest_mock.MockerFixture, package_test_no_backend: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_no_backend)

    assert builder._build_system == {'requires': [], 'build-backend': DEFAULT_BACKEND['build-backend']}


def test_missing_requires(mocker: pytest_mock.MockerFixture, package_test_no_requires: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    with pytest.raises(build.BuildException):
        build.ProjectBuilder(package_test_no_requires)


def test_build_system_typo(mocker: pytest_mock.MockerFixture, package_test_typo: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    with pytest.warns(build.TypoWarning):
        build.ProjectBuilder(package_test_typo)


def test_missing_outdir(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)
    builder._hook.build_sdist.return_value = 'dist.tar.gz'  # type: ignore[attr-defined]
    out = os.path.join(tmp_dir, 'out')

    builder.build('sdist', out)

    assert os.path.isdir(out)


def test_relative_outdir(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:  # noqa: ARG001
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)
    builder._hook.build_sdist.return_value = 'dist.tar.gz'  # type: ignore[attr-defined]

    builder.build('sdist', '.')

    builder._hook.build_sdist.assert_called_with(os.path.abspath('.'), None)  # type: ignore[attr-defined]


def test_build_not_dir_outdir(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)
    builder._hook.build_sdist.return_value = 'dist.tar.gz'  # type: ignore[attr-defined]
    out = os.path.join(tmp_dir, 'out')

    open(out, 'a', encoding='utf-8').close()  # create empty file

    with pytest.raises(build.BuildException):
        builder.build('sdist', out)


@pytest.fixture(scope='session')
def demo_pkg_inline(tmp_path_factory: pytest.TempPathFactory) -> pathlib.Path:
    # builds a wheel without any dependencies and with a console script demo-pkg-inline
    tmp_path = tmp_path_factory.mktemp('demo-pkg-inline')
    builder = build.ProjectBuilder(source_dir=os.path.join(os.path.dirname(__file__), 'packages', 'inline'))
    out = tmp_path / 'dist'
    builder.build('wheel', str(out))
    return next(out.iterdir())


@pytest.mark.contextvars
@pytest.mark.isolated
def test_build_with_dep_on_console_script(
    tmp_path: pathlib.Path, demo_pkg_inline: pathlib.Path, capfd: pytest.CaptureFixture[str], mocker: pytest_mock.MockerFixture
) -> None:
    """All command-line scripts provided by the build-required packages must be present in the build environment's PATH."""
    # we first install demo pkg inline as build dependency (as this provides a console script we can check)
    # to validate backend invocations contain the correct path we use an inline backend that will fail, but first
    # provides the PATH information (and validates shutil.which is able to discover the executable - as PEP states)
    toml = textwrap.dedent(
        """
        [build-system]
        requires = ["demo_pkg_inline"]
        build-backend = "build"
        backend-path = ["."]

        [project]
        description = "Factory ⸻ A code generator 🏭"
        authors = [{name = "Łukasz Langa"}]
        """
    )
    code = textwrap.dedent(
        """
        import os
        import shutil
        import sys
        print("BB " + os.environ["PATH"])
        exe_at = shutil.which("demo-pkg-inline")
        print("BB " + exe_at)
        """
    )
    (tmp_path / 'pyproject.toml').write_text(toml, encoding='UTF-8')
    (tmp_path / 'build.py').write_text(code, encoding='utf-8')

    deps = {str(demo_pkg_inline)}  # we patch the requires demo_pkg_inline to refer to the wheel -> we don't need index
    mocker.patch('build.ProjectBuilder.build_system_requires', new_callable=mocker.PropertyMock, return_value=deps)
    from build.__main__ import main

    with pytest.raises(SystemExit):
        main(['--wheel', '--outdir', str(tmp_path / 'dist'), str(tmp_path)])

    out, _ = capfd.readouterr()
    lines = [line[3:] for line in out.splitlines() if line.startswith('BB ')]  # filter for our markers
    path_vars = lines[0].split(os.pathsep)
    which_detected = lines[1]
    assert which_detected.startswith(path_vars[0]), out


def test_prepare(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)
    builder._hook.prepare_metadata_for_build_wheel.return_value = 'dist-1.0.dist-info'  # type: ignore[attr-defined]

    assert builder.prepare('wheel', tmp_dir) == os.path.join(tmp_dir, 'dist-1.0.dist-info')
    builder._hook.prepare_metadata_for_build_wheel.assert_called_with(tmp_dir, None, _allow_fallback=False)  # type: ignore[attr-defined]


def test_prepare_no_hook(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)
    failure = pyproject_hooks.HookMissing('prepare_metadata_for_build_wheel')
    builder._hook.prepare_metadata_for_build_wheel.side_effect = failure  # type: ignore[attr-defined]

    assert builder.prepare('wheel', tmp_dir) is None


def test_prepare_error(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)
    builder._hook.prepare_metadata_for_build_wheel.side_effect = Exception  # type: ignore[attr-defined]

    with pytest.raises(build.BuildBackendException, match='Backend operation failed: Exception'):
        builder.prepare('wheel', tmp_dir)


def test_prepare_not_dir_outdir(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)

    builder = build.ProjectBuilder(package_test_flit)

    out = os.path.join(tmp_dir, 'out')
    with open(out, 'w', encoding='utf-8') as f:
        f.write('Not a directory')
    with pytest.raises(build.BuildException, match=r'Build path .* exists and is not a directory'):
        builder.prepare('wheel', out)


def test_no_outdir_single(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller.prepare_metadata_for_build_wheel', return_value='')

    builder = build.ProjectBuilder(package_test_flit)

    out = os.path.join(tmp_dir, 'out')
    builder.prepare('wheel', out)

    assert os.path.isdir(out)


def test_no_outdir_multiple(mocker: pytest_mock.MockerFixture, tmp_dir: str, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller.prepare_metadata_for_build_wheel', return_value='')

    builder = build.ProjectBuilder(package_test_flit)

    out = os.path.join(tmp_dir, 'does', 'not', 'exist')
    builder.prepare('wheel', out)

    assert os.path.isdir(out)


def test_runner_user_specified(tmp_dir: str, package_test_flit: str) -> None:
    def dummy_runner(cmd: typing.Sequence[str], cwd: str | None = None, extra_environ: dict[str, str] | None = None) -> None:  # noqa: ARG001
        msg = 'Runner was called'
        raise RuntimeError(msg)

    builder = build.ProjectBuilder(package_test_flit, runner=dummy_runner)  # type: ignore[arg-type]
    with pytest.raises(build.BuildBackendException, match='Runner was called'):
        builder.build('wheel', tmp_dir)


def test_metadata_path_no_prepare(tmp_dir: str, package_test_no_prepare: str) -> None:
    builder = build.ProjectBuilder(package_test_no_prepare)

    metadata = _importlib.metadata.PathDistribution(
        pathlib.Path(builder.metadata_path(tmp_dir)),
    ).metadata
    assert metadata is not None

    assert metadata['name'] == 'test-no-prepare'
    assert metadata['Version'] == '1.0.0'


def test_metadata_path_with_prepare(tmp_dir: str, package_test_setuptools: str) -> None:
    builder = build.ProjectBuilder(package_test_setuptools)

    metadata = _importlib.metadata.PathDistribution(
        pathlib.Path(builder.metadata_path(tmp_dir)),
    ).metadata
    assert metadata is not None

    # Setuptools < v69.0.3 (https://github.com/pypa/setuptools/pull/4159) normalized this to dashes
    assert metadata['name'].replace('-', '_') == 'test_setuptools'
    assert metadata['Version'] == '1.0.0'


def test_metadata_path_legacy(tmp_dir: str, package_legacy: str) -> None:
    builder = build.ProjectBuilder(package_legacy)

    metadata = _importlib.metadata.PathDistribution(
        pathlib.Path(builder.metadata_path(tmp_dir)),
    ).metadata
    assert metadata is not None

    assert metadata['name'] == 'legacy'
    assert metadata['Version'] == '1.0.0'


def test_metadata_invalid_wheel(tmp_dir: str, package_test_bad_wheel: str) -> None:
    builder = build.ProjectBuilder(package_test_bad_wheel)

    with pytest.raises(ValueError, match='Invalid wheel'):
        builder.metadata_path(tmp_dir)


def test_log(mocker: pytest_mock.MockerFixture, caplog: pytest.LogCaptureFixture, package_test_flit: str) -> None:
    mocker.patch('pyproject_hooks.BuildBackendHookCaller', autospec=True)
    mocker.patch('build.ProjectBuilder._call_backend', return_value='some_path')
    caplog.set_level(logging.DEBUG)

    builder = build.ProjectBuilder(package_test_flit)
    builder.get_requires_for_build('sdist')
    builder.get_requires_for_build('wheel')
    builder.prepare('wheel', '.')
    builder.build('sdist', '.')
    builder.build('wheel', '.')

    assert [(record.levelname, record.message) for record in caplog.records] == [
        ('INFO', 'Getting build dependencies for sdist...'),
        ('INFO', 'Getting build dependencies for wheel...'),
        ('INFO', 'Getting metadata for wheel...'),
        ('INFO', 'Building sdist...'),
        ('INFO', 'Building wheel...'),
    ]


@pytest.mark.parametrize(
    ('pyproject_toml', 'parse_output'),
    [
        (
            {'build-system': {'requires': ['foo']}},
            {'requires': ['foo'], 'build-backend': 'setuptools.build_meta:__legacy__'},
        ),
        (
            {'build-system': {'requires': ['foo'], 'build-backend': 'bar'}},
            {'requires': ['foo'], 'build-backend': 'bar'},
        ),
        (
            {'build-system': {'requires': ['foo'], 'build-backend': 'bar', 'backend-path': ['baz']}},
            {'requires': ['foo'], 'build-backend': 'bar', 'backend-path': ['baz']},
        ),
    ],
)
def test_parse_valid_build_system_table_type(pyproject_toml: dict[str, Any], parse_output: dict[str, Any]) -> None:
    assert build._builder._parse_build_system_table(pyproject_toml) == parse_output


@pytest.mark.parametrize(
    ('pyproject_toml', 'error_message'),
    [
        (
            {'build-system': {}},
            '`requires` is a required property',
        ),
        (
            {'build-system': {'requires': 'not an array'}},
            '`requires` must be an array of strings',
        ),
        (
            {'build-system': {'requires': [1]}},
            '`requires` must be an array of strings',
        ),
        (
            {'build-system': {'requires': ['foo'], 'build-backend': ['not a string']}},
            '`build-backend` must be a string',
        ),
        (
            {'build-system': {'requires': ['foo'], 'backend-path': 'not an array'}},
            '`backend-path` must be an array of strings',
        ),
        (
            {'build-system': {'requires': ['foo'], 'backend-path': [1]}},
            '`backend-path` must be an array of strings',
        ),
        (
            {'build-system': {'requires': ['foo'], 'unknown-prop': False}},
            'Unknown properties: unknown-prop',
        ),
    ],
)
def test_parse_invalid_build_system_table_type(pyproject_toml: dict[str, Any], error_message: str) -> None:
    with pytest.raises(build.BuildSystemTableValidationError, match=error_message):
        build._builder._parse_build_system_table(pyproject_toml)


@pytest.mark.parametrize(
    'setup',
    [
        pytest.param(lambda _tmp_path: None, id='nonexistent'),
        pytest.param(lambda tmp_path: (tmp_path / 'bad').write_text('', encoding='utf-8'), id='file-not-dir'),
    ],
)
def test_backend_path_invalid_directory(tmp_path: pathlib.Path, setup: Callable[[pathlib.Path], None]) -> None:
    (tmp_path / 'pyproject.toml').write_text(
        textwrap.dedent("""\
            [build-system]
            requires = []
            build-backend = "backend"
            backend-path = ["bad"]
        """),
        encoding='utf-8',
    )
    setup(tmp_path)
    with pytest.raises(build.BuildSystemTableValidationError, match='does not exist or is not a directory'):
        build.ProjectBuilder(tmp_path)



from __future__ import annotations


__lazy_modules__ = ['tomli', 'tomllib']

import sys


if sys.version_info >= (3, 11):
    from tomllib import TOMLDecodeError, load, loads  # pragma: no cover
else:  # pragma: no cover
    from tomli import TOMLDecodeError, load, loads


__all__ = [
    'TOMLDecodeError',
    'load',
    'loads',
]
